#!/usr/bin/env python3

"""Use the hotcorner feature to execute custom logic.

This script responds to hotcorner events, which are activated when the pointer
reaches any of the eight hotcorner edges [see TODO]. These events will always
trigger, regardless of whether tiling is enabled. This serves as a basic
template, useful as a starting point if you only wish to use the
hotcorner features without the tiling functionality of cortile.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile

Usage:
    Run the python application::

        $ python use_hotcorners_without_tiling.py

"""


from cortile import Cortile
from cortile.helper.dict import Dict


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # listen to cortile events, the lambda function just passes the cortile and event object
    ct.listen(lambda event: event_callback(ct, event))

    # this prevents the main method from exiting and therefore keeps the script running
    ct.wait()


def event_callback(ct: Cortile, event: Dict):

    # the callback will fire on all events, here we are only interested on corner events
    if event.Name == 'Corner':
        handle_hotcorner_activation(ct, event)


def handle_hotcorner_activation(ct: Cortile, event: Dict):

    # receive the corner event data
    corner = event.Data
    if not corner:
        return

    # print the activated corner name, desktop and screen location
    print(f'Corner: {corner.Name}', corner.Location)

    # TODO: run some custom python logic
    if corner.Name == 'top_left':
        pass
    if corner.Name == 'top_center':
        pass
    if corner.Name == 'top_right':
        pass
    if corner.Name == 'center_right':
        pass
    if corner.Name == 'bottom_right':
        pass
    if corner.Name == 'bottom_center':
        pass
    if corner.Name == 'bottom_left':
        pass
    if corner.Name == 'center_left':
        pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Apply a layout and disable tiling afterwards.

You can use this script in cases where you only want to maximize
windows once and disable tiling afterwards. The active workspace and
screen index is determined by the current position of the mouse. At first
tiling will be enabled, then maximization will be applied [see TODO].
Afterwards tiling is disabled and windows are in stacking mode again.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile

Usage:
    Run the python application::

        $ python apply_layout_and_disable.py

"""


from cortile import Cortile


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # retrieve current active desktop/workspace index
    desktop_index = ct.get_active_workspace()

    # retrieve current active screen index
    screen_index = ct.get_active_screen()

    # enable tiling, just in case it was disabled (since cortile actions will be otherwise ignored)
    ct.action_execute_enable(desktop=desktop_index, screen=screen_index)

    # TODO: change to your desired layout
    ct.action_execute_layout_maximized(desktop=desktop_index, screen=screen_index)

    # disable tiling, but leave all windows at there latest position
    ct.action_execute_disable(desktop=desktop_index, screen=screen_index)

    # this closes the connection to the running cortile process gracefully
    ct.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Make a specific application always master.

When cortile is running and a pre-defined application is launched,
it will make the application a master window on the current layout.
The application will be selected based on the class name(s), which
are passed as command line arguments [see TODO]. Applications
will be made master only if there where just recently opened.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile

Usage:
    Run the python application (TODO: pass class names of applications).
    Within the cortile 'addons' folder this script will not work, since
    arguments can't be passed.::

        $ python make_app_always_master.py firefox code

"""

import sys
from typing import List
from cortile import Cortile
from cortile.helper.dict import Dict
from cortile.helper.time import Time


# helper variable to store the last event time
event_time = Time()


def main(args):

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # listen to cortile events, the lambda function just passes the cortile, event and args object
    ct.listen(lambda event: event_callback(ct, event, args))

    # this prevents the main method from exiting and therefore keeps the script running
    ct.wait()


def event_callback(ct: Cortile, event: Dict, args: List[str]):

    # the callback will fire on all events, here we are only interested on client changes
    if event.Name == 'Clients':
        handle_clients_change(ct, event, args)


def handle_clients_change(ct: Cortile, event: Dict, args: List[str]):

    # to prevent multiple executions, we’ll ignore additional client events triggered by our own logic
    if event_time.delta().seconds < 1:
        return

    # windows classes passed as list of argument strings
    window_classes = [x.lower() for x in args]

    # the event data values contains a list with all tracked cortile clients
    for client in event.Data.Values:

        # x11 window id as integer, other applications may provide the id as hex value 0x...
        window_id = client.Window.Id

        # x11 window class name, based on this we can filter all application instances of this type
        window_class = client.Latest.Class.lower()

        # cortile location (not to be confused with the geometry) of the window, holds the desktop and screen index
        window_location = client.Latest.Location

        # cortile creation timestamp in ms, is the time at which the client was tracked internally
        window_timestamp = client.Window.Created

        # timedelta relative to now, to determine whether the window has just been recently opened
        window_lifetime = Time(window_timestamp).delta()

        # lets check if one of the clients match and they have just been recently (in the last 3s) added
        if window_class in window_classes and window_lifetime.seconds < 3:

            # this moves the focus to the window
            ct.window_activate(id=window_id)

            # the focused window will be made a master, the desktop/workspace and screen index must be provided
            ct.action_execute_master_make(desktop=window_location.Desktop, screen=window_location.Screen)

            # update the global event time object
            event_time.set(event.Time)


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python3

"""Swap the active window with the window you are pointing at.

This example demonstrates how to move and swap windows within the active
workspace. A swap event is triggered only if the source and target windows
are on the same desktop and screen. The primary goal of this script is to
illustrate how to access pointer device data and swap windows by setting the
target position [see TODO], which can be any position on the active workspace.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile

Usage:
    Run the python application::

        $ python swap_apps_using_pointer.py

"""


from cortile import Cortile
from cortile.helper.dict import Dict


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # listen to cortile events, the lambda function just passes the cortile and event object
    ct.listen(lambda event: event_callback(ct, event))

    # this prevents the main method from exiting and therefore keeps the script running
    ct.wait()


def event_callback(ct: Cortile, event: Dict):

    # the callback will fire on all events, here we are only interested on pointer events
    if event.Name == 'Pointer':
        handle_pointer_clicks(ct, event)


def handle_pointer_clicks(ct: Cortile, event: Dict):

    # receive the pointer event data
    pointer = event.Data
    if not pointer:
        return

    # extract the clicked device button and position on the screen
    button = next((k for k, v in pointer.Device.Button.items() if v), None)
    position = pointer.Device.Position

    # print the clicked device button and position on the screen
    print(f'Pointer: {button}', position)

    # extract the active client and window id
    client = ct.get_active_client()
    if not client:
        return
    window_id = client.Window.Id

    # TODO: move the focused window to the middle click position, which will swap windows
    if button == 'Middle':
        ct.window_to_position(id=window_id, x=position.X, y=position.Y)
    if button == 'Left':
        pass
    if button == 'Right':
        pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import sys
import signal

from types import FrameType


class Signal(object):
    def __init__(self):
        """
        Initialize the signal handler.
        This helper class manages sigint and sigterm events to signal whether
        an application exit has been requested, ensuring a gracefully termination.
        """
        self.sigcount = 0
        signal.signal(signal.SIGINT, self.sig)
        signal.signal(signal.SIGTERM, self.sig)

    def sig(self, number: int, frame: FrameType) -> None:
        """
        Handle signal events.

        :param number: Signal number, not used
        :param frame: Signal frame, not used
        """
        self.sigcount += 1
        if self.sigcount > 2:
            sys.exit(0)

    @property
    def exit(self) -> bool:
        """
        Flag that indicates if sig() event was triggered.

        :return: True if sig event was triggered, False otherwise
        """
        return self.sigcount > 0



#!/usr/bin/env python3

import os
import sys
import syslog

from datetime import datetime

from cortile.helper.dict import Dict


class Logger(object):

    LEVELS = Dict(
        DEBUG=0,
        INFO=1,
        WARN=2,
        ERROR=3,
        FATAL=4
    )

    COLORS = Dict(
        RED='\033[31m',
        GREEN='\033[32m',
        BLUE='\033[34m',
        YELLOW='\033[33m',
        MAGENTA='\033[35m',
        CYAN='\033[36m',
        GRAY='\033[37m',
        WHITE='\033[97m',
        BLACK='\033[30m',
        RESET='\033[0m'
    )

    def __init__(self, level: int = 2):
        """
        Initialize the syslog logger.
        This helper class writes logging message to stdout and syslog, where
        outputs to stdout are highlighted by colors, depending on the severity.

        :param level: Logging level, default is warn
        """
        self.level = level
        self.pid = os.getpid()

    def color(self, level: str) -> str:
        """
        Get color value for logging level.

        :param level: Logging level as string

        :return: Color value of logging level
        """
        return Dict(
            DEBUG=self.COLORS.GRAY,
            INFO=self.COLORS.CYAN,
            WARN=self.COLORS.YELLOW,
            ERROR=self.COLORS.MAGENTA,
            FATAL=self.COLORS.RED
        )[level]

    def prefix(self, level: str, colored: bool = True) -> str:
        """
        Get prefix header for log message.

        :param level: Logging level as string
        :param colored: Flag indicating color usage

        :return: Logging prefix string
        """
        time = f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}'
        pid = f'{self.pid}'
        context = f'{level}'
        if colored:
            time = f'{self.COLORS.GRAY}{time}{self.COLORS.RESET}'
            pid = f'{self.COLORS.GRAY}{pid}{self.COLORS.RESET}'
            context = f'{self.color(level)}{context}{self.COLORS.RESET}'
        return f'{time} | {pid} | {context}'

    def log(self, level: str, text: str) -> None:
        """
        Write message to stdout and syslog.

        :param level: Logging level as string
        :param text: Logging text as string
        """
        if not (self.LEVELS[level] >= self.level):
            return
        print(f'{self.prefix(level, True)} | {text}')
        syslog.syslog(f'{self.prefix(level, False)} | {text}'.replace('\n', ' '))

    def debug(self, text: str) -> None:
        """
        Write debug message.

        :param text: Logging text as string
        """
        self.log('DEBUG', text)

    def info(self, text: str) -> None:
        """
        Write info message.

        :param text: Logging text as string
        """
        self.log('INFO', text)

    def warn(self, text: str) -> None:
        """
        Write warn message.

        :param text: Logging text as string
        """
        self.log('WARN', text)

    def error(self, text: str) -> None:
        """
        Write error message.

        :param text: Logging text as string
        """
        self.log('ERROR', text)

    def fatal(self, text: str) -> None:
        """
        Write fatal message and exit process.

        :param text: Logging text as string
        """
        self.log('FATAL', text)
        sys.exit(1)



#!/usr/bin/env python3

import time

from typing import Callable, Iterator, List

from cortile.helper.dict import Dict
from cortile.helper.logger import Logger
from cortile.base.connector import Connector


class Cortile(object):
    def __init__(self, log: int = Logger.LEVELS.WARN):
        """
        Initialize the cortile connector.
        This main class wraps methods of the base connector and should be
        used as primary interface to communicate with a running cortile instance.

        :param log: Logging level, default is warn
        """
        self.connector = Connector(log)

    @property
    def log(self) -> Logger:
        """
        Return the logger instance.

        :return: Logger instance that writes to syslog
        """
        return self.connector.log

    def listen(self, callback: Callable[[Dict], None] | None) -> None:
        """
        Start listening for events.

        :param callback: Function to call when an event is received
        """
        self.connector.listen(callback)

    def wait(self, sleep: float = 0.5) -> None:
        """
        Keeps the process running for the connector to listen.

        :param sleep: Time to sleep in between, default is 0.5 seconds
        """
        while self.connector.connected and not self.connector.exit:
            time.sleep(sleep)
        self.close()

    def close(self) -> None:
        """
        Close the connection gracefully.
        """
        self.connector.close()

    def get_active_layout(self) -> Dict | None:
        """
        Get the active layout for the current desktop and screen.

        :return: Active layout with tiling enabled or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return None
        for layout in self.get_active_layouts():
            if layout.Location.Desktop == workplace.CurrentDesktop and layout.Location.Screen == workplace.CurrentScreen:
                return layout
        return None

    def get_active_layouts(self) -> Iterator[Dict]:
        """
        Get the active layouts from the workspaces.

        :return: Iterator of active layouts with tiling enabled
        """
        workspaces = self.connector.property('Workspaces')
        if not workspaces:
            return
        for workspace in workspaces.Values:
            if workspace.Tiling:
                yield workspace.Layouts[workspace.Layout]
        return

    def get_active_client(self) -> Dict | None:
        """
        Get the current focused client window.

        :return: Active client or None
        """
        clients = self.get_clients()
        windows = self.get_windows()
        for client in clients:
            if windows and windows.Active.Id == client.Window.Id:
                return client
        return None

    def get_active_clients(self) -> Iterator[Dict]:
        """
        Get information of clients on the current active screen.

        :return: Iterator of tracked clients on the current screen
        """
        clients = self.connector.property('Clients')
        workplace = self.connector.property('Workplace')
        if not clients or not workplace:
            return
        for client in clients.Values:
            location = client.Latest.Location
            if location.Desktop == workplace.CurrentDesktop and location.Screen == workplace.CurrentScreen:
                yield client
        return

    def get_active_workspace(self) -> int | None:
        """
        Get the current active workspace.

        :return: Active workspace index or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return None
        return workplace.CurrentDesktop

    def get_active_screen(self) -> int | None:
        """
        Get the current active screen.

        :return: Active screen index or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return None
        return workplace.CurrentScreen

    def get_workspace_count(self) -> int | None:
        """
        Get the number of workspaces.

        :return: Number of workspaces or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return None
        return workplace.DesktopCount

    def get_screen_count(self) -> int | None:
        """
        Get the number of screens.

        :return: Number of screens or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return None
        return workplace.ScreenCount

    def get_workspace_dimensions(self) -> List[Dict]:
        """
        Get the dimensions of all workspaces.

        :return: LTR sorted list of workspace dimensions or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return []
        return workplace.Displays.Desktops

    def get_screen_dimensions(self) -> List[Dict]:
        """
        Get the dimensions of all screens.

        :return: LTR sorted list of screen dimensions or None
        """
        workplace = self.connector.property('Workplace')
        if not workplace:
            return []
        return workplace.Displays.Screens

    def get_clients(self) -> List[Dict]:
        """
        Get all the clients information.

        :return: List of tracked clients or None
        """
        clients = self.connector.property('Clients')
        if not clients:
            return []
        return clients.Values

    def get_windows(self) -> Dict | None:
        """
        Get all the windows information.

        :return: List of tracked window ids or None
        """
        windows = self.connector.property('Windows')
        if not windows:
            return None
        return windows

    def desktop_switch(self, desktop: int) -> bool:
        """
        Switch to a different desktop.

        :param desktop: Index of the desktop to switch to

        :return: True if successful, False otherwise
        """
        return self.connector.method('DesktopSwitch', desktop)

    def window_activate(self, id: int) -> bool:
        """
        Activate a window by its id.

        :param id: Id of the window to activate

        :return: True if successful, False otherwise
        """
        return self.connector.method('WindowActivate', id)

    def window_to_desktop(self, id: int, desktop: int) -> bool:
        """
        Move a window to a different desktop.

        :param id: Id of the window to move
        :param desktop: Index of the desktop to move the window to

        :return: True if successful, False otherwise
        """
        return self.connector.method('WindowToDesktop', id, desktop)

    def window_to_position(self, id: int, x: int, y: int) -> bool:
        """
        Move a window to a specific position.

        :param id: Id of the window to move
        :param x: X coordinate to move the window to
        :param y: Y coordinate to move the window to

        :return: True if successful, False otherwise
        """
        return self.connector.method('WindowToPosition', id, x, y)

    def window_to_screen(self, id: int, screen: int) -> bool:
        """
        Move a window to a different screen.

        :param id: Id of the window to move
        :param screen: Index of the screen to move the window to

        :return: True if successful, False otherwise
        """
        return self.connector.method('WindowToScreen', id, screen)

    def action_execute_enable(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'enable' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'enable', desktop, screen)

    def action_execute_disable(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'disable' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'disable', desktop, screen)

    def action_execute_toggle(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'toggle' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'toggle', desktop, screen)

    def action_execute_decoration(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'decoration' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'decoration', desktop, screen)

    def action_execute_restore(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'restore' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'restore', desktop, screen)

    def action_execute_cycle_next(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'cycle_next' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'cycle_next', desktop, screen)

    def action_execute_cycle_previous(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'cycle_previous' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'cycle_previous', desktop, screen)

    def action_execute_layout_vertical_left(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'layout_vertical_left' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'layout_vertical_left', desktop, screen)

    def action_execute_layout_vertical_right(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'layout_vertical_right' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'layout_vertical_right', desktop, screen)

    def action_execute_layout_horizontal_top(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'layout_horizontal_top' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'layout_horizontal_top', desktop, screen)

    def action_execute_layout_horizontal_bottom(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'layout_horizontal_bottom' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'layout_horizontal_bottom', desktop, screen)

    def action_execute_layout_maximized(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'layout_maximized' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'layout_maximized', desktop, screen)

    def action_execute_layout_fullscreen(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'layout_fullscreen' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'layout_fullscreen', desktop, screen)

    def action_execute_master_make(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'master_make' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'master_make', desktop, screen)

    def action_execute_master_make_next(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'master_make_next' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'master_make_next', desktop, screen)

    def action_execute_master_make_previous(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'master_make_previous' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'master_make_previous', desktop, screen)

    def action_execute_master_increase(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'master_increase' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'master_increase', desktop, screen)

    def action_execute_master_decrease(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'master_decrease' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'master_decrease', desktop, screen)

    def action_execute_slave_increase(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'slave_increase' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'slave_increase', desktop, screen)

    def action_execute_slave_decrease(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'slave_decrease' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'slave_decrease', desktop, screen)

    def action_execute_proportion_increase(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'proportion_increase' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'proportion_increase', desktop, screen)

    def action_execute_proportion_decrease(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'proportion_decrease' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'proportion_decrease', desktop, screen)

    def action_execute_window_next(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'window_next' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'window_next', desktop, screen)

    def action_execute_window_previous(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'window_previous' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'window_previous', desktop, screen)

    def action_execute_reset(self, desktop: int, screen: int) -> bool:
        """
        Execute the 'reset' action.

        :param desktop: Index of the desktop
        :param screen: Index of the screen

        :return: True if successful, False otherwise
        """
        return self.connector.method('ActionExecute', 'reset', desktop, screen)



#!/usr/bin/env python3

"""Automatically open applications on specific workspace.

When cortile is running and a pre-defined application is launched,
it will automatically relocate the application to a designated workspace.
This behavior is determined by the applications class name(s). Both the class
name(s) and the target workspace are hardcoded into the system [see TODO]. This
ensures that each application is consistently moved to its appropriate workspace.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile

Usage:
    Run the python application::

        $ python move_app_to_workspace.py

"""


from cortile import Cortile
from cortile.helper.dict import Dict
from cortile.helper.time import Time


# helper variable to store the last event time
event_time = Time()


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # listen to cortile events, the lambda function just passes the cortile and event object
    ct.listen(lambda event: event_callback(ct, event))

    # this prevents the main method from exiting and therefore keeps the script running
    ct.wait()


def event_callback(ct: Cortile, event: Dict):

    # the callback will fire on all events, here we are only interested on client changes
    if event.Name == 'Clients':
        handle_clients_change(ct, event)


def handle_clients_change(ct: Cortile, event: Dict):

    # to prevent multiple executions, we’ll ignore additional client events triggered by our own logic
    if event_time.delta().seconds < 1:
        return

    # TODO: change desktop index and window classes to your preferences
    desktop_index = 1
    window_classes = ['firefox']

    # the event data values contains a list with all tracked cortile clients
    for client in event.Data.Values:

        # x11 window id as integer, other applications may provide the id as hex value 0x...
        window_id = client.Window.Id

        # x11 window class name, based on this we can filter all application instances of this type
        window_class = client.Latest.Class

        # cortile location (not to be confused with the geometry) of the window, holds the desktop and screen index
        window_location = client.Latest.Location

        # cortile creation timestamp in ms, is the time at which the client was tracked internally
        window_timestamp = client.Window.Created

        # timedelta in relation to now, to determine whether the window has just been opened recently
        window_lifetime = Time(window_timestamp).delta()

        # lets check if one of the clients match and they have just been recently (in the last 3s) added
        if window_class in window_classes and window_lifetime.seconds < 3:

            # move the window to the pre-defined desktop/workspace index
            if window_location.Desktop != desktop_index:
                ct.window_to_desktop(id=window_id, desktop=desktop_index)

                # update the global event time object
                event_time.set(event.Time)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import subprocess

from threading import Thread, Event
from typing import Callable, Tuple, IO


class Process(Thread):
    def __init__(self, *args: Tuple[str, ...]):
        """
        Initialize the process thread.
        This base class runs a subprocess in a background thread and returns
        stdout and stderr, either synchronously or asynchronously using callbacks.

        :param args: Process binary path and arguments
        """
        super().__init__(daemon=True)
        self.process = None
        self.callback = None
        self.open = Event()
        if len(args):
            self.process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.open.set()

    @property
    def running(self) -> bool:
        """
        Flag that indicates if subprocess is still running.

        :return: True if subprocess is running, False otherwise
        """
        return self.open.is_set()

    def run(self) -> None:
        """
        Run the subprocess and send stdout to callback.
        """
        if not isinstance(self.process, subprocess.Popen):
            return
        while callable(self.callback) and self.process.poll() is None:
            try:
                stdout = self.process.stdout.readline()
                if stdout is None or len(stdout) == 0:
                    continue
                self.callback(stdout, b'', 0)
            except Exception as e:
                break
        self.open.clear()

    def communicate(self, callback: Callable[[IO, IO, int], None] | None = None) -> Tuple[IO, IO, int] | None:
        """
        Communicate synchronous or asynchronous with subprocess.

        :param callback: Optional callback function for asynchronous calls
        """
        if not isinstance(self.process, subprocess.Popen):
            return
        if not callable(callback):
            return *self.process.communicate(), self.process.poll()
        self.callback = callback
        self.start()

    def terminate(self) -> None:
        """
        Terminate the subprocess thread gracefully.
        """
        if not isinstance(self.process, subprocess.Popen):
            return
        self.process.terminate()



#!/usr/bin/env python3

"""Open a python gui window showing a list of clients.

This example opens a python gui built with tkinter to display a list of clients present
on the currently active screen. Each client is represented by a simple button, which when
clicked, triggers a pre-defined cortile event. In this case [see TODO], the event will
make the window a master. The script demonstrates how to retrieve screen geometry,
which is used here to position the gui window at the center of the screen.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ apt install python3-tk
        $ pip install cortile

Usage:
    Run the python application::

        $ python open_gui_with_clients.py

"""


import tkinter as tk
from cortile import Cortile


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # retrieve the screen dimension of the currently active screen
    screens = ct.get_screen_dimensions()
    screen_index = ct.get_active_screen()
    screen_dimension = screens[screen_index].Geometry

    # calculate the tkinter window position
    window_w = 400
    window_h = 400
    window_x = screen_dimension.X + screen_dimension.Width // 2 - window_w // 2
    window_y = screen_dimension.Y + screen_dimension.Height // 2 - window_h // 2

    # create a tkinter window positioned at the center of the screen
    window = tk.Tk()
    window.title('cortile-addons')
    window.geometry(f'{window_w}x{window_h}+{window_x}+{window_y}')
    window.overrideredirect(True)
    window.resizable(False, False)
    window.configure(bg='white')

    # tkinter button to close the window
    tk.Button(window, text='[CLOSE GUI]', command=window.destroy).pack()

    # retrieve clients on the current active desktop and screen
    for client in ct.get_active_clients():
        window_id = client.Window.Id
        client_name = client.Latest.Name
        client_class = client.Latest.Class

        # TODO: tkinter button to make window master
        button_text = f'{client_class} - {client_name}'[:40]
        tk.Button(window, text=f'{button_text}...', command=function_factory(ct, window_id)).pack()

    # start the tkinter event loop
    window.mainloop()

    # this closes the connection to the running cortile process gracefully
    ct.close()


def function_factory(ct: Cortile, window_id: int):

    # wrapper function to be attached on hotkey events
    return lambda: master_make(ct, window_id)


def master_make(ct: Cortile, window_id: int):

    # retrieve current active desktop and screen index
    active_desk = ct.get_active_workspace()
    active_screen = ct.get_active_screen()

    # activate the clicked client and make it a master window
    ct.window_activate(id=window_id)
    ct.action_execute_master_make(desktop=active_desk, screen=active_screen)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Bind global hotkeys and apply a layout.

This script is designed for one-time window arrangement followed
by disabling tiling. Keyboard shortcuts are registered directly within
the python script [see TODO], therefore bypassing cortile key bindings.
The script will wait for global hotkey events, which executes python
callback functions that will trigger your desired cortile logic.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile pynput

Usage:
    Run the python application::

        $ python apply_layout_with_keybindings.py

"""


from cortile import Cortile
from pynput import keyboard


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # TODO: bind keyboard shortcuts and callback functions
    with keyboard.GlobalHotKeys({
        '<ctrl>+<alt>+i': lambda: key_callback_i(ct),
        '<ctrl>+<alt>+j': lambda: key_callback_j(ct)
    }) as hotkey:
        hotkey.join()

    # this closes the connection to the running cortile process gracefully
    ct.close()


def key_callback_i(ct: Cortile):
    print('<ctrl>+<alt>+i pressed')

    # apply the horizontal-top layout
    apply_layout(ct, 'horizontal_top')


def key_callback_j(ct: Cortile):
    print('<ctrl>+<alt>+j pressed')

    # apply the vertical-right layout
    apply_layout(ct, 'vertical_right')


def apply_layout(ct: Cortile, name: str):

    # retrieve current active desktop/workspace index
    desktop_index = ct.get_active_workspace()

    # retrieve current active screen index
    screen_index = ct.get_active_screen()

    # enable tiling, just in case it was disabled (since cortile actions will be otherwise ignored)
    ct.action_execute_enable(desktop=desktop_index, screen=screen_index)

    # apply the selected layout
    if name == 'horizontal_top':
        ct.action_execute_layout_horizontal_top(desktop=desktop_index, screen=screen_index)
    elif name == 'vertical_right':
        ct.action_execute_layout_vertical_right(desktop=desktop_index, screen=screen_index)

    # disable tiling, but leave all windows at there latest position
    ct.action_execute_disable(desktop=desktop_index, screen=screen_index)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import os
import time
import dbus

from typing import Callable, Tuple, IO

from cortile.helper.dict import Dict
from cortile.base.process import Process


class Session(object):
    def __init__(self, name: str = 'com.github.leukipp.cortile', path: str = '/com/github/leukipp/cortile'):
        """
        Initialize the dbus connector.
        This base class connects to the running cortile instance and communicates
        with the build in cortile dbus client. The results are stored into dictionaries.

        :param name: Dbus name, default is com.github.leukipp.cortile
        :param path: Dbus path, default is /com/github/leukipp/cortile
        """
        self.name = name
        self.path = path
        self.file = str()

    @property
    def connected(self) -> bool:
        """
        Flag that indicates if connect() was successful.

        :return: True if cortile binary exists, False otherwise
        """
        return os.path.exists(self.file)

    def connect(self) -> Dict:
        """
        Retrieve cortile binary path with a direct dbus call.

        :return: Dictionary with success or error data
        """
        try:
            self.file = str(dbus.SessionBus().get_object(self.name, self.path).Get(self.name, 'Process')['Path'])
        except Exception as e:
            return self.data('Error', Message=repr(e))
        return self.data('Result', Success=True)

    def disconnect(self) -> None:
        """
        Disconnect session by resetting cortile binary path.
        """
        self.file = str()

    def listen(self, callback: Callable[[Dict], None], *args: Tuple[str, ...]) -> Process:
        """
        Listen asynchronously to cortile events.

        :param callback: Callback function for cortile action events
        :param args: Optional arguments to filter cortile event types

        :return: Running or empty background process thread
        """
        if not self.connected:
            if callable(callback):
                callback(self.data('Error', Message='Not connected'))
            return Process()
        process = Process(self.file, 'dbus', '-listen', *map(str, args))
        process.communicate(lambda a, b, c: callback(self.parse(a, b, c)))
        return process

    def method(self, name: str, *args: Tuple[str, ...]) -> Dict:
        """
        Execute cortile method with arguments.

        :param name: Name of the cortile method
        :param args: Arguments of the cortile method

        :return: Dictionary with success or error data
        """
        if not self.connected:
            return self.data('Error', Message='Not connected')
        process = Process(self.file, 'dbus', '-method', name, *map(str, args))
        return self.parse(*process.communicate())

    def property(self, name: str) -> Dict:
        """
        Retrieve cortile property.

        :param name: Name of the cortile property

        :return: Dictionary with success or error data
        """
        if not self.connected:
            return self.data('Error', Message='Not connected')
        process = Process(self.file, 'dbus', '-property', name)
        return self.parse(*process.communicate())

    def help(self) -> Dict:
        """
        Show the help message from cortile dbus -help.

        :return: Dictionary with success or error data
        """
        if not self.connected:
            return self.data('Error', Message='Not connected')
        process = Process(self.file, 'dbus', '-help')
        return self.parse(*process.communicate())

    @staticmethod
    def parse(stdout: IO, stderr: IO, code: int) -> Dict:
        """
        Parse stdout and stderr messages from subprocess.

        :param stdout: Success output of subprocess
        :param stderr: Error output of subprocess
        :param code: Status code of subprocess

        :return: Dictionary with success or error data
        """
        out = stdout.decode('utf-8').strip()
        err = stderr.decode('utf-8').strip()
        if out.startswith('{') and out.endswith('}'):
            return Dict.from_json(out)
        return Session.data('Error', Message=f'{out} {err} {"(" + str(code) + ")" if code else ""}'.strip())

    @staticmethod
    def data(typ: str, **kwargs: dict[str, object]) -> Dict:
        """
        Create formatted data dict which matches cortile return structure.

        :param typ: Type of the data
        :param kwargs: Keyword arguments of the data

        :return: Dictionary with success or error data
        """
        return Dict(
            Process=os.getpid(),
            Time=int(time.time_ns() / 1e6),
            Type=typ,
            Name='Connector',
            Data=Dict(kwargs)
        )



#!/usr/bin/env python3

from typing import Callable, Tuple

from cortile.helper.dict import Dict
from cortile.helper.logger import Logger
from cortile.helper.signal import Signal
from cortile.base.session import Session


class Connector(object):
    def __init__(self, log: int = Logger.LEVELS.WARN):
        """
        Initialize the session connector.
        This base class acts as a middle layer and wraps session methods for
        registration of listener callbacks and caching of received cortile properties.

        :param log: Logging level, default is warn
        """
        self.log = Logger(log)
        self.signal = Signal()
        self.session = Session()
        self.properties = Dict()
        self.listener = [self.observe]
        result = self.session.connect()
        if result.Type == 'Result' and result.Data.Success:
            self.log.info('Init: Connection established')
        if result.Type == 'Error':
            self.log.fatal(f'Error: {result.Data.Message}')
        self.process = self.session.listen(self.callbacks)

    @property
    def connected(self) -> bool:
        """
        Flag that indicates if session.connect() was successful.

        :return: True if cortile binary is running, False otherwise
        """
        return self.process.running and self.session.connected

    @property
    def exit(self) -> bool:
        """
        Flag that indicates if signal.sig() event was triggered.

        :return: True if sig event was triggered, False otherwise
        """
        return self.signal.exit

    def close(self) -> None:
        """
        Close the connection gracefully.
        """
        self.log.info(f'Close connection: {self.session.file}')
        self.session.disconnect()
        self.process.terminate()

    def listen(self, callback: Callable[[Dict], None]) -> None:
        """
        Listen asynchronously to cortile events.

        :param callback: Callback function for cortile action events
        """
        self.log.info(f'Register listener: {len(self.listener)}')
        self.listener.append(callback)

    def method(self, name: str, *args: Tuple[str, ...]) -> bool:
        """
        Execute cortile method with arguments.

        :param name: Name of the cortile method
        :param args: Arguments of the cortile method

        :return: True if successful, False otherwise
        """
        self.log.info(f'Method: {name} {" ".join(map(str, args))}')
        result = self.session.method(name, *args)
        if result.Type == 'Error':
            self.log.error(f'Error: {result.Data.Message}')
        return result.Type == 'Result' and result.Data.Success

    def property(self, name: str, cached: bool = True) -> Dict | None:
        """
        Retrieve cortile property.

        :param name: Name of the cortile property
        :param cached: Use the cached value from internal listener, default is True

        :return: Dictionary with success data or None on failure
        """
        self.log.info(f'Property: {name}')
        if name not in self.properties or not cached:
            result = self.session.property(name)
            if result.Type == 'Error':
                self.log.error(f'Error: {result.Data.Message}')
            if result.Type == 'Property':
                self.properties[name] = result.Data
        return self.properties[name] if name in self.properties else None

    def help(self) -> str:
        """
        Show the help message from cortile dbus -help.

        :return: String with help message output
        """
        return self.session.help().Data.Message

    def observe(self, result: Dict | None) -> None:
        """
        Internal function to update cached properties or disconnect client.

        :param result: Dictionary with success or error data
        """
        if not result or result.Type != 'Property':
            return
        if result.Name == 'Disconnect':
            return self.close()
        self.log.info(f'{result.Type}: {result.Name} update')
        self.properties[result.Name] = result.Data

    def callbacks(self, result: Dict | None) -> None:
        """
        Internal function to execute registered callback functions.

        :param result: Dictionary with success or error data
        """
        if not result:
            return
        for callback in self.listener:
            if not callable(callback):
                continue
            callback(result)



#!/usr/bin/env python3

import json


class Dict(dict):
    def __init__(self, *args: tuple[object, ...], **kwargs: dict[str, object]):
        """
        Initialize a dot notation dictionary.
        This helper class is used globally to provide simplified dictionary
        access by using the dot notation for attribute setter and getter methods.

        :param args: Dictionary argument
        :param kwargs: Keyword arguments
        """
        super().__init__(*args, **kwargs)
        for arg in args:
            if not isinstance(arg, dict):
                continue
            for k, v in arg.items():
                if isinstance(v, dict):
                    self[k] = Dict(v)
                elif isinstance(v, list):
                    self[k] = [Dict(x) if isinstance(x, dict) else x for x in v]
                else:
                    self[k] = v
        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    @staticmethod
    def from_json(string: str) -> object:
        """
        Instantiate a dot notation dictionary from json string.

        :param string: Dictionary as json string

        :return: Dot notation dictionary instance
        """
        return Dict(json.loads(string))

    def __getattr__(self, key: str) -> object:
        """
        Get dot notation dictionary attribute.

        :param key: Dictionary attribute key

        :return: Dictionary attribute value
        """
        return self.__getitem__(key)

    def __setattr__(self, key: str, value: object) -> None:
        """
        Set dot notation dictionary attribute.

        :param attr: Dictionary attribute key
        :param attr: Dictionary attribute value
        """
        self.__setitem__(key, value)

    def __delattr__(self, key: str) -> None:
        """
        Delete dot notation dictionary attribute.

        :param attr: Dictionary attribute key
        """
        self.__delitem__(key)

    def __str__(self) -> str:
        """
        Serialize dot notation dictionary as string.

        :return: Dictionary as json string
        """
        return json.dumps(self, indent=2)



#!/usr/bin/env python3

from datetime import datetime, timezone, timedelta


class Time(object):
    def __init__(self, ts: int = 0.0, tz: timezone = timezone.utc):
        """
        Initialize the time helper.
        This helper class converts unix timestamps in milliseconds into datetime objects.
        Additionally, it can calculate the time difference relative to the current moment.

        :param ts: Initial unix timestamp in milliseconds, default is 0.0
        :param tz: Optional datetime timezone, default is utc
        """
        self.tz = tz
        self.dt = datetime.fromtimestamp(ts / 1000.0, tz=self.tz)

    def get(self) -> datetime:
        """
        Get datetime object from stored time.

        :return: Datetime object obtained from timestamp
        """
        return self.dt

    def set(self, ts: int) -> None:
        """
        Set datetime object from unix timestamp.

        :param ts: Unix timestamps in milliseconds
        """
        self.dt = datetime.fromtimestamp(ts / 1000.0, tz=self.tz)

    def now(self) -> datetime:
        """
        Get datetime object with current time.

        :return: Datetime object of now
        """
        return datetime.now(tz=self.tz)

    def delta(self) -> timedelta:
        """
        Get time difference to the stored datetime object.

        :return: Timedelta between now and stored time
        """
        return self.now() - self.dt



#!/usr/bin/env python3

"""Export the current tiling state into files.

This example listens to cortile events and handles workspace updates.
It retrieves the current layouts and exports tiling states and decoration
states for each screen and desktop into files. The export folder path is
hardcoded [see TODO]. This script serves as an example of how to retrieve
properties from active layouts and to use it afterward for custom logic.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile

Usage:
    Run the python application::

        $ python write_tiling_state_file.py

"""


import os
from cortile import Cortile
from cortile.helper.dict import Dict


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # listen to cortile events, the lambda function just passes the cortile and event object
    ct.listen(lambda event: event_callback(ct, event))

    # this prevents the main method from exiting and therefore keeps the script running
    ct.wait()


def event_callback(ct: Cortile, event: Dict):

    # the callback will fire on all events, here we are only interested on workspace updates
    if event.Name == 'Workspaces':
        handle_workspace_update(ct)


def handle_workspace_update(ct: Cortile):

    # TODO: adjust export folder path
    export_folder = os.path.join(os.path.sep, 'tmp', 'cortile')

    # create folder structure
    if not os.path.exists(export_folder):
        os.makedirs(export_folder)

    # dictionary that stores a mapping between file path and file contents
    files = Dict()

    # initialize dictionary with default values, key is file path and value the "active_layout_name;decoration_state"
    for screen_index in range(ct.get_screen_count()):
        for workspace_index in range(ct.get_workspace_count()):
            file_name = f'screen{screen_index}-desktop{workspace_index}.txt'
            file_path = os.path.join(export_folder, file_name)

            # initialize files dictionary
            files[file_path] = 'disabled;None'

    # update dictionary with actual values, get_active_layouts() only returns layouts with tiling enabled
    for layout in ct.get_active_layouts():
        file_name = f'screen{layout.Location.Screen}-desktop{layout.Location.Desktop}.txt'
        file_path = os.path.join(export_folder, file_name)

        # update files dictionary
        files[file_path] = f'{layout.Name};{layout.Decoration}'

    # export state files
    export_state_files(files)


def export_state_files(files: Dict):
    print(f'Export files: {files}')

    # iterate over the dictionary and write states to the file system
    for file_path, file_content in files.items():
        with open(file_path, 'w') as file:
            file.write(file_content)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Move app to another screen with global hotkeys.

This script binds hotkeys in the format <ctrl>+<alt>+[SCREEN_INDEX], where SCREEN_INDEX
represents the index of one of the connected screens [see TODO]. The screens are ordered
from left to right (LTR) based on their physical arrangement. When a hotkey is pressed,
the currently active (focused) window will move to the target screen on the same
desktop (workspace). Tiling must be enabled on the source and target screens.

Authors:
    * https://github.com/leukipp/

Dependencies:
    The following packages have to be installed::

        $ pip install cortile pynput

Usage:
    Run the python application::

        $ python move_app_to_screen.py

"""


from cortile import Cortile
from pynput import keyboard


def main():

    # init a cortile python object and connect to the running cortile process
    ct = Cortile()

    # TODO: map hotkeys to screen index
    hotkeys = dict()
    for i in range(ct.get_screen_count()):
        hotkeys[f'<ctrl>+<alt>+{i}'] = function_factory(ct, i)

    # bind keyboard shortcuts and callback functions
    with keyboard.GlobalHotKeys(hotkeys) as hotkey:
        hotkey.join()

    # this closes the connection to the running cortile process gracefully
    ct.close()


def function_factory(ct: Cortile, screen_index: int):

    # wrapper function to be attached on hotkey events
    return lambda: move_to_screen(ct, screen_index)


def move_to_screen(ct: Cortile, screen_index: int):

    # extract the active client and window id
    client = ct.get_active_client()
    if not client:
        return

    # x11 window id as integer, other applications may provide the id as hex value 0x...
    window_id = client.Window.Id

    # cortile location (not to be confused with the geometry) of the window, holds the desktop and screen index
    window_location = client.Latest.Location

    # move client window to screen
    if window_location.Screen != screen_index:
        ct.window_to_screen(id=window_id, screen=screen_index)


if __name__ == '__main__':
    main()
